<?php

namespace jrosset\Collections;

use ArrayAccess;
use Closure;
use Countable;
use IteratorAggregate;
use JsonSerializable;
use OutOfBoundsException;
use Serializable;
use Throwable;

/**
 * Interface for an immutable (read-only) collection
 *
 * @template TKey of array-key
 * @template TValue
 *
 * @implements IteratorAggregate<TKey, TValue>
 * @implements ArrayAccess<TKey, TValue>
 * @implements IArrayCast<TKey, TValue>
 */
interface IImmutableCollection extends IteratorAggregate, JsonSerializable, Serializable, Countable, ArrayAccess, IArrayCast {
    /**
     * Checks if the collection is empty
     *
     * @return bool TRUE if the collection is empty, FALSE otherwise
     */
    public function empty (): bool;

    /**
     * Checks if a key exists
     *
     * @param TKey $key The key
     *
     * @return bool TRUE if the key exists, FALSE otherwise
     */
    public function exists ($key): bool;
    /**
     * Checks if contains a value
     *
     * @param TValue $value  The value
     * @param bool   $strict Strict comparison ?
     *
     * @return bool TRUE if the value exists, FALSE otherwise
     */
    public function contains ($value, bool $strict = false): bool;

    /**
     * Get the value of a key or null if not found
     *
     * @param TKey $key The key
     *
     * @return TValue|null The value
     */
    public function get ($key);
    /**
     * Get the value at an offset
     *
     * @param int $offset The offset
     *
     * @return TValue The value
     *
     * @throws OutOfBoundsException If the offset is not valid
     */
    public function getFromOffset (int $offset);

    /**
     * Get the first key of a value or null if not found
     *
     * @param TValue $value  The value
     * @param bool   $strict Strict comparison ?
     *
     * @return TKey|null
     */
    public function key ($value, bool $strict = false);
    /**
     * Get the key at an offset
     *
     * @param int $offset The offset
     *
     * @return TKey The key
     *
     * @throws OutOfBoundsException If the offset is not valid
     */
    public function keyFromOffset (int $offset);

    /**
     * Get the value of the first element
     *
     * @return TValue The value
     *
     * @throws OutOfBoundsException If the collection is empty
     */
    public function first ();
    /**
     * Get the key of the first element
     *
     * @return TKey The key
     *
     * @throws OutOfBoundsException If the collection is empty
     */
    public function firstKey ();

    /**
     * Get the value of the last element
     *
     * @return TValue The value
     *
     * @throws OutOfBoundsException If the collection is empty
     */
    public function last ();
    /**
     * Get the key of the last element
     *
     * @return TKey The key
     *
     * @throws OutOfBoundsException If the collection is empty
     */
    public function lastKey ();

    /**
     * Extract a slice of the collection
     *
     * Preserves the keys
     *
     * @param int      $offset The start offset
     * @param int|null $length The maximum length. Null if until end of the collection
     *
     * @return static<TKey, TValue> The result collection
     *
     * @throws OutOfBoundsException If the offset is not valid
     *
     * @noinspection PhpMissingReturnTypeInspection
     */
    public function slice (int $offset, ?int $length = null);

    /**
     * Get a collection of all elements that satisfy predicate $filter
     *
     * @param Closure(TKey, TValue):bool $filter The filtering predicate
     *
     * @return static<TKey, TValue> The result collection
     *
     * @noinspection PhpMissingReturnTypeInspection
     */
    public function filter (Closure $filter);
    /**
     * Get a collection of all not empty elements
     *
     * Use {@see https://www.php.net/manual/function.empty.php empty} to check if element is empty or not
     *
     * @return static<TKey, TValue> The result collection
     *
     * @noinspection PhpMissingReturnTypeInspection
     */
    public function withoutEmpties ();
    /**
     * A new collection with <b>$process</b> applied on all elements
     *
     * @template TResultValue
     *
     * @param Closure(TKey, TValue): TResultValue $process The process function to apply on each element
     *
     * @return static<TKey, TResultValue> The result collection
     *
     * @noinspection PhpMissingReturnTypeInspection
     */
    public function map (Closure $process);

    /**
     * Get a collection with the elements sorted (by value)
     *
     * @param null|IComparator|Closure(TValue, TValue):int $sorter The sorting method ; Null if values are object implementing {@see IComparable}.
     *                                                             <br>Return :
     *                                                             <br>&lt; 0 if value1 is before value2
     *                                                             <br>= 0 if value1 equals value2
     *                                                             <br>&gt; 0 if value1 is after value2
     *
     * @return static<TKey, TValue> The result collection
     *
     * @throws Throwable If an error occurs
     *
     * @noinspection PhpMissingReturnTypeInspection
     */
    public function sort ($sorter = null);
    /**
     * Get a collection with the elements sorted by key
     *
     * @param IComparator|Closure(TKey, TKey):int $sorter The sorting method. Return :
     *                                                    <br>&lt; 0 if key1 is before key2
     *                                                    <br>= 0 if key1 equals key2
     *                                                    <br>&gt; 0 if key1 is after key2
     *
     * @return static<TKey, TValue> The result collection
     *
     * @throws Throwable If an error occurs
     *
     * @noinspection PhpMissingReturnTypeInspection
     */
    public function sortByKey ($sorter);

    /**
     * The list of all keys
     *
     * @return IImmutableCollection<int, TKey> The list of all keys
     */
    public function keys (): IImmutableCollection;
    /**
     * The list of all values
     *
     * @return static<int, TValue> The list of all values
     *
     * @noinspection PhpMissingReturnTypeInspection
     */
    public function values ();

    /**
     * Join all values with a separator
     *
     * @param string $separator The separator
     *
     * @return string The joined values
     */
    public function join (string $separator): string;
    /**
     * Create a collection of split values
     *
     * @param string $value     The value to split
     * @param string $separator The split separator
     *
     * @return static<int, string> The result collection
     *
     * @noinspection PhpMissingReturnTypeInspection
     */
    public static function split (string $value, string $separator);

    /**
     * Create a collection filled with the same element
     *
     * @param int    $size  The collection size
     * @param TValue $value The value to fill the collection
     *
     * @return static<int, TValue> The result collection
     *
     * @noinspection PhpMissingReturnTypeInspection
     */
    public static function fill (int $size, $value);
    /**
     * Create a collection filled with the same element, using a list of keys
     *
     * @param IImmutableCollection<TKey> $keys  The list of keys
     * @param TValue                     $value The value to fill the collection
     *
     * @return static<TKey, TValue> The result collection
     *
     * @noinspection PhpMissingReturnTypeInspection
     */
    public static function fillWithKeys (IImmutableCollection $keys, $value);

    /**
     * Merge the current collection with one or multiple collections
     *
     * @param IImmutableCollection<TKey, TValue> ...$collections The other collections to merge
     *
     * @return static<TKey, TValue> The result collection
     *
     * @noinspection PhpMissingReturnTypeInspection
     */
    public function mergeWith (IImmutableCollection ...$collections);
}